# -*- coding: utf-8 -*-
"""
Rest endpoint base implementation.

Endpoints take care of interacting with HTTP. They for a bridge between HTTP
requests/responses and the REST resources. They should never be implemented
manually and ideally most apps should not need to subclass those but instead
implement all of their logic in the `RestResource` subclasses.

The code here is just an abstract base that holds the code common to all
endpoints. It is not self sufficient, but only minor web-framework dependent
code is left for the subclasses.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
from collections import namedtuple
from logging import getLogger

# local imports
from .resource import RestResource
from . import params


L = getLogger(__name__)
RestResult = namedtuple('RestResult', 'status,headers,data')


class RestEndpoint(object):
    """ Abstract base class for all REST endpoints.

    The endpoint responsibility is to provide a bridge between HTTP and
    `RestResource`.  The latter should never directly deal with HTTP.

    """

    def __init__(self, res_cls):
        if self.is_resource(res_cls):
            raise ValueError('res_cls must be a subclass of RestResource')

        self.resource = res_cls()

    def is_resource(self, res_cls):
        """ Check if the given object is a resource class.

        :param type res_cls:
            Object to test.
        :return bool:
            **True** if *res_cls* is a resource class.
        """
        return not (isinstance(res_cls, type) and
                    issubclass(res_cls, RestResource))

    def get_ok_status(self, verb):
        """ Get a valid ok status code for a given rest verb. """
        if verb == 'create':
            return 201
        elif verb == 'delete':
            return 204
        else:
            return 200

    def process_result(self, result, default_status):
        """ Unify the handler result into a stable format.

        The unified format is ``(status, headers, data)``.

        :param tuple|dict|list result:
            The result as returned from the handler
        :param int default_status:
            Default HTTP status to use (if result contains none).
        :return tuple(int, dict, dict):
            Tuple . The handlers can return data in
            the shortened forms, either ``(status, data)`` or just ``data``.
            This method will expand those results with default and return a
            unified result tuple.
        """
        status = default_status
        headers = {}
        data = None

        if isinstance(result, tuple):
            if len(result) == 2:
                status, data = result
            elif len(result) == 3:
                status, headers, data = result
        else:
            data = result

        return RestResult(status, headers, data)

    def call_rest_handler(self, method, request):
        """

        :param str method:
            HTTP method. Must be uppercase
        :param request:
            HTTP request. It is assumed the request object hast the following
            members: ``rest_keys``,
        :return tuple(int, dict, dict):
            Tuple ``(status, headers, data)``. The handlers can return data in
            the shortened forms, either ``(status, data)`` or just ``data``.
            This method will expand those results with default and return a
            unified result tuple.
        """
        my_pk = self.resource.get_pk(request)

        rest_verb = self.determine_rest_verb(method, my_pk)
        if rest_verb is None:
            return RestResult(405, {}, None)

        handler = getattr(self.resource, 'rest_' + rest_verb, None)
        if handler is None or not callable(handler):
            return RestResult(405, {}, None)

        try:
            handler_args = self.build_handler_args(rest_verb, request)
            result = handler(request, **handler_args)
            return self.process_result(result, self.get_ok_status(rest_verb))

        except ValueError as ex:
            L.exception('Invalid REST invocation')
            L.error(ex)
            return RestResult(400, {}, {'detail': str(ex)})

        except NotImplementedError as ex:
            return RestResult(405, {}, {'detail': str(ex)})

        except Exception as ex:
            L.exception('Unhandled resource exception')
            L.error(ex)
            return RestResult(500, {}, {'detail': str(ex)})

    @classmethod
    def determine_rest_verb(cls, http_method, pk):
        """ Return the REST operation associated with the given HTTP method.

        This is part of the `RestEndpoint` class because this way it becomes
        customisable by subclasses.

        :param str|unicode http_method:
            The HTTP method used for the request.
        :param str|int pk:
            The primary key extracted from the URL template.
        :return str|unicode:
            The REST verb associated with that request. This is also the name
            of the method on `RestResource` that will be called by the endpoint.
        """
        if http_method == 'GET':
            return 'query' if pk is None else 'get'
        elif http_method == 'POST' and pk is None:
            return 'create'
        elif http_method == 'PUT' and pk is not None:
            return 'update'
        elif http_method == 'DELETE' and pk is not None:
            return 'delete'
        elif http_method == 'OPTIONS':
            return 'options'
        elif http_method == 'HEAD':
            return 'head'
        else:
            return None

    @classmethod
    def extract_request_data(cls, request):
        """ Extract request data as a python data structure

        This method is here to provide an easy way to implement new framework
        support. The way the request data is handled is framework dependant
        so each endpoint implementation for the given framework must provide
        and implementation of this function.

        :param Request request:
            Underlying framework dependant request.
        :return:
            A python data structure (dict,list,string, primitive type) that is
            the request data extract from the underlying
        """
        raise NotImplementedError(
            "RestEndpoint sublasses must implement .extract_request_data()"
        )

    @classmethod
    def extract_request_query_string(cls, request):
        """ Extract request params as a python dictionary.

        This method is here to provide an easy way to implement new framework
        support. The way the request params are handled is framework dependant
        so each endpoint implementation for the given framework must provide
        an implementation of this function.

        :param Request request:
            Underlying framework dependant request.
        :return:
            A python dict with ``{name: value}`` pairs for all request query
            string params
        """
        return request.GET

    @classmethod
    def build_handler_args(cls, verb, request):
        """ Build handler invocation arguments.

        The parameters extracted here can be directly passed to the
        `RestResource` handler method for the given REST *verb*.

        This is part of the `RestEndpoint` class because this way it becomes
        customisable by subclasses.

        :param str|unicode verb:
            REST verb/action. This is what the `RestResource` classes abstract.
            Can be one of ``get/query/create/update/delete/options/head``.
        :param HttpRequest request:
            An HTTP request the handler will serve. Handler args are extracted
            from it here.
        :param None|int|str|unicode pk:
            The primary key of a resource that will be modified by request or
            ``None`` if the request is to list handlers. This is most likely
            extracted from the URL template and that's why it's not passed as
            part of the request.
        :return dict:
        """
        handler_args = {}

        if verb in ('create', 'update'):
            handler_args['data'] = cls.extract_request_data(request)

        if verb in ('query', 'get'):
            qs = cls.extract_request_query_string(request)
            handler_args['params'] = params.parse(qs)

        return handler_args



# -*- coding: utf-8 -*-
""" Flask integration. """
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json
from logging import getLogger
from urllib.parse import urljoin

# 3rd party imports
from flask import request
from restible import RestEndpoint

# local imports
from restible import api_action, api_route


L = getLogger(__name__)


class FlaskEndpoint(RestEndpoint):
    """ Endpoint implementation to use in webapp2/AppEngine projects. """
    def __init__(self, res_cls, protected=False):
        super(FlaskEndpoint, self).__init__(res_cls)
        self.protected = protected

    @classmethod
    def extract_request_data(cls, request):
        return request.json

    @classmethod
    def extract_request_query_string(cls, request):
        return request.args

    def authorize(self, request):
        """ Authorize user from request. """
        raise NotImplementedError("Subclasses must implement .authorize()")

    def dispatch(self, **params):
        """ Override webapp2 dispatcher. """
        request.rest_keys = params

        request.user = self.authorize(request)
        if self.protected and request.user is None:
            return {'detail': "Not authorized"}, 401

        result = self.call_rest_handler(request.method, request)

        return json.dumps(result.data), result.status, result.headers

    def find_action(self, name, generic):
        """ Find action by name and type. """
        for action in self.resource.rest_actions:
            meta = api_action.get_meta(action)

            if meta.name == name and meta.generic == generic:
                return action

        return None

    def call_action(self, action, payload):
        """ Call API action. """
        action_params = {'payload': payload}
        meta = api_action.get_meta(action)

        if not meta.generic:
            obj = self.resource.get_requested(request)

            if obj is None:
                pk = self.resource.get_pk(request)
                return 404, {
                    'detail': "{} #{} not found".format(self.resource.name, pk)
                }

            action_params['obj'] = obj

        L.info("Calling action {}".format(action))
        return action(**action_params)

    def dispatch_action(self, name, generic, **params):
        """ Override webapp2 dispatcher. """
        request.rest_keys = params

        action = self.find_action(name, generic)
        if action is None:
            return {
                'detail': "action {} not found on {}".format(
                    name, self.resource.name
                )
            }, 401

        meta = api_action.get_meta(action)

        request.user = self.authorize(request)
        if meta.protected and request.user is None:
            return {"detail": "Not authorized"}, 401

        result = self.call_action(action, request.json)
        result = self.process_result(result, 200)

        return json.dumps(result.data), result.status, result.headers

    @classmethod
    def init_app(cls, app, resources=None, routes=None):
        """ Initialize the endpoint configuration for the given flask app. """
        resources = resources or []
        routes = routes or []
        cls.endpoints = []

        for entry in resources:
            opts = {}

            if len(entry) == 2:
                url, res_cls = entry
            else:
                url, res_cls, opts = entry

            if not url.endswith('/'):
                url += '/'

            endpoint = cls(res_cls, **opts)
            endpoint._register_routes(app, url)

            cls.endpoints.append(endpoint)

        for url, route in routes:
            meta = api_route.get_meta(route)
            app.add_url_rule(url[:-1], view_func=route, **meta)
            app.add_url_rule(url, view_func=route, **meta)

    def _register_routes(self, app, url):
        """ Register all routes for the current endpoint in the flask app. """
        url_list = url
        url_item = urljoin(url, '<{name}_pk>/'.format(
            name=self.resource.name
        ))

        # actions
        for action in self.resource.rest_actions():
            meta = api_action.get_meta(action)

            action_route = dict(
                endpoint='{}-{}'.format(self.resource.name, meta.name),
                defaults={
                    'name': meta.name,
                    'generic': meta.generic,
                },
                view_func=self.dispatch_action,
                methods=['post']
            )

            base_url = url_list if meta.generic else url_item
            url_action = urljoin(base_url, meta.name + '/')

            # Add without and with slash
            app.add_url_rule(url_action[:-1], **action_route)
            app.add_url_rule(url_action, **action_route)

        item_route = dict(
            endpoint='{}-item'.format(self.resource.name),
            view_func=self.dispatch,
            methods=['get', 'put', 'delete']
        )
        list_route = dict(
            endpoint='{}-list'.format(self.resource.name),
            view_func=self.dispatch,
            methods=['get', 'post']
        )

        # Add without and with slash
        app.add_url_rule(url_item[:-1], **item_route)
        app.add_url_rule(url_item, **item_route)

        # Add without and with slash
        app.add_url_rule(url_list[:-1], **list_route)
        app.add_url_rule(url_list, **list_route)



# -*- coding: utf-8 -*-
"""
Django implementation of a `RestEndpoint`.
"""
from __future__ import absolute_import, unicode_literals

# stdlib imports
import json

# 3rd party imports
from django.conf.urls import url
from django.http import JsonResponse
from six import iteritems

# local imports
from .. import RestEndpoint


class DjangoEndpoint(RestEndpoint):
    """ Endpoint implementation to use in django projects. """
    
    def dispatch(self, request, **keys):
        """ Django 'calls' views so this is what will be ran by django.

        This should cover the whole request lifecycle.
        """
        request.rest_keys = keys

        result = self.call_rest_handler(request.method, request)

        response = JsonResponse(
            status=result.status,
            safe=False,
            data=result.data
        )
        for key, value in iteritems(result.headers):
            response[key] = value

        return response

    @classmethod
    def extract_request_data(cls, request):
        try:
            if request.body:
                body = request.body.decode('utf-8')
                return json.loads(body)
            else:
                return None
        except:
            raise ValueError("Invalid JSON body: '{}'".format(
                request.body.decode('utf-8')
            ))

    @property
    def urls(self):
        """ Return all the URLs required by the endpoint.

        :return list<url>:
            A list of URLs that can be directly appended to ``urlpatterns`` of
            choice.
        """
        # TODO: Using resource.name for both prefix and PK name might blow up.
        # There might be one generic endpoint called /bars and we might want to
        # create a nested one /foo/bars. In those cases we have two resources
        # with the same name - bars.
        #
        # Another question is whether this is an issue. In the above example
        # the resources are not at the same 'level' in the tree. This will
        # definitely fail if we have a recursive endpoint, but that might be
        # even worse.
        name = self.resource.name
        ret = [
            url(
                '{name}/(?P<{name}_pk>[^/]+)/?$'.format(name=name),
                self.dispatch,
                name='{name}-detail'.format(name=name)
            ),
            url(
                '{name}/?$'.format(name=name),
                self.dispatch,
                name='{name}-list'.format(name=name)
            ),
        ]
        return ret
